/*  DeltaSMT: Delta-debugger for Satisfiablity Modulo Theories (SMT) benchmarks.
 *  Copyright (C) 2009  Robert Daniel Brummayer
 *
 *  This file is part of DeltaSMT.
 *
 *  DeltaSMT is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DeltaSMT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.util.*;

public class DeltaSMT
{

  protected String inName;

  protected String outName;

  protected String tmpName;
  
  protected Runtime rt;

  protected String []cmdArray;

  protected int goldenExitCode;

  protected int verbosity;

  protected boolean normSharing;

  private long timeoutMillis;

  private static String strSuccess;

  private static String strFail;

  private static String strTimeout;

  private static final int offset = 65;


  public DeltaSMT(String inName, String outName, String tmpName, 
                  String []cmdArray, int verbosity, boolean colors,
    long timeoutMillis, boolean normSharing){
    assert (inName != null);
    assert (outName != null);
    assert (tmpName != null);
    assert (cmdArray != null);
    assert (verbosity >= 0);
    assert (timeoutMillis >= 1000l);

    this.inName = inName;
    this.outName = outName;
    this.tmpName = tmpName;
    this.rt = Runtime.getRuntime();
    this.cmdArray = new String[cmdArray.length + 1];
    for (int i = 0; i < cmdArray.length; i++)
      this.cmdArray[i] = cmdArray[i];
    this.cmdArray[this.cmdArray.length - 1] = tmpName;
    this.verbosity = verbosity;
    this.timeoutMillis = timeoutMillis;
    this.normSharing = normSharing;

    if (colors) {
      this.strSuccess = (char) 27 + "[1;34m[" + (char) 27 + "[0;32mSUCCESS" +
                        (char) 27 + "[1;34m]" + (char) 27 + "[0;39m";
      this.strFail = (char) 27 + "[1;34m[" + (char) 27 + "[0;31mFAILED" + 
                     (char) 27 + "[1;34m]" + (char) 27 + "[0;39m";
      this.strTimeout = (char) 27 + "[1;34m[" + (char) 27 + "[1;30mTIMEOUT" +
                        (char) 27 + "[1;34m]" + (char) 27 + "[0;39m";
    } else {
      this.strSuccess = "[SUCCESS]";
      this.strFail = "[FAILED]";
      this.strTimeout = "[TIMEOUT]";
    }
  }

  protected static void copyFile (String dest, String src) throws IOException {
    BufferedReader in;
    PrintStream out;
    String s;

    assert (dest != null);
    assert (src != null);

    in = new BufferedReader (new FileReader(src));
    out = new PrintStream (dest);
    while ((s = in.readLine()) != null)
      out.println (s);
    in.close();
    out.close();
  }

  protected void printlnVerbosityMsg (String msg, int verbosity) {
    if (this.verbosity >= verbosity) {
      System.out.println (msg);
      System.out.flush ();
    }
  }

  protected void printVerbosityMsg (String msg, int verbosity) {
    if (this.verbosity >= verbosity)
      System.out.print (msg);
  }

  protected void writeSimplifiedFormula (SMTBenchmark bench,
                                         HashMap<Long, SMTNode> subst)
    throws IOException {
    PrintStream tmpStream;

    assert (bench != null);
    assert (subst != null);

    tmpStream = new PrintStream(this.tmpName);
    bench.dumpSubstitute (tmpStream, subst);
    tmpStream.close();
  }

  protected SimplifiedResult runOnSimplifiedFormula () throws  IOException {
    SimplifiedResult result;
    int exitCode;
    Timer timer;
    Process proc;

    result = SimplifiedResult.NO;
    timer = new Timer();
    proc = this.rt.exec(this.cmdArray);
    timer.schedule (new Interrupter (Thread.currentThread()), 
                    this.timeoutMillis);
    try {
      exitCode = proc.waitFor();
      if (exitCode == this.goldenExitCode)
        result = SimplifiedResult.YES;
    } catch (InterruptedException ie) {
      result = SimplifiedResult.TIMEOUT;
    } finally {
      proc.destroy();
      timer.cancel();
    }
    return result;
  }

  protected SMTBenchmark checkSimplifiedFormula (SMTNodeManager mgr, 
                                                 SMTBenchmark bench,
                                                 HashMap<Long, SMTNode> subst, 
                                                 Queue<SMTNode> queue) 
    throws InterruptedException, IOException, ParserException {
    SMTBenchmark result = null;
    Stack<SMTNode> newNodes;
    SMTNode node;

    assert (mgr != null);
    assert (bench != null);
    assert (subst != null);

    writeSimplifiedFormula (bench, subst);
    switch (runOnSimplifiedFormula ()){
      case YES:
        printlnVerbosityMsg (strSuccess, 3);
        copyFile (this.outName, this.tmpName);
        node = subst.get (bench.getRootFormula().getId());
        mgr.startNodeRecording();
        Parser.init(mgr, new BufferedReader (new FileReader(this.tmpName)));
        result = Parser.parseSMTBenchmark();
        Parser.reset();
        newNodes = mgr.collectRecordedNodesAndStopRecording();
        while (!newNodes.empty())
          queue.add (newNodes.pop());
        break;
      case NO:
        printlnVerbosityMsg (strFail, 3);
        break;
      case TIMEOUT:
        printlnVerbosityMsg (strTimeout, 3);
        break;
    }
    return result;
  }

  protected static long getFileSize (String fileName) throws IOException {
    File file;
    assert (fileName != null);
    file = new File (fileName);
    return file.length();
  }

  public void deltaDebug() throws ParserException, InterruptedException, 
    IOException {

    String msg;
    Process proc;
    long origSize = 0;
    long reducedSize = 0;
    boolean simplified;
    SMTNode []children;
    Queue<SMTNode> queue;
    SMTNode cur, zero, one, tmp;
    SMTNode root = null;
    SMTBenchmark newBench;
    SMTBenchmark bench = null;
    HashMap<Long, SMTNode> skip, subst;
    HashSet<SMTNode> cache = null;
    SMTNode []formulas;
    List<SMTNode> assumptions;
    SMTNodeManager mgr = new SMTNodeManager (new SMTTypeManager(),
                                             this.normSharing);

    /* copy input file to tmp file */
    copyFile (tmpName, this.inName);
    copyFile (this.outName, this.inName);

    if (this.verbosity > 0) {
      printVerbosityMsg ("file size of original formula: ", 1);
      origSize = getFileSize (this.inName);
      printlnVerbosityMsg (origSize + " bytes", 1);
    }

    printVerbosityMsg ("computing golden exit code: ", 1);

    proc = rt.exec(this.cmdArray);
    this.goldenExitCode = proc.waitFor();
    printlnVerbosityMsg(String.valueOf(this.goldenExitCode), 1);

    printlnVerbosityMsg  ("parsing original formula", 1);
    Parser.init(mgr, new BufferedReader (new FileReader(this.inName)));
    bench = Parser.parseSMTBenchmark();
    Parser.reset();

    /* setting status of benchmark to 'unknown' as status is likely to
     * change during delta-debugging */
    if (bench.getStatus() != null)
      bench.setStatus (SMTBenchmarkStatus.UNKNOWN);

    subst = new HashMap<Long,SMTNode>();
    queue = new LinkedList<SMTNode>();

    assumptions = bench.getAssumptions();
    if (assumptions != null && assumptions.size() > 0) {
      printlnVerbosityMsg ("found assumption(s)", 0);
      printlnVerbosityMsg ("trying to merge assumption(s) into main formula", 
                           0);

      formulas = new SMTNode[assumptions.size() + 1];
      for (int i = 0; i < assumptions.size(); i++)
        formulas[i] = assumptions.get(i);
      formulas[assumptions.size()] = bench.getRootFormula();
      try {
        root = mgr.and (formulas);
      } catch (TypeException te) {
        System.err.println("INTERNAL ERROR");
        System.exit(1);
      }
      mgr.release (bench.getRootFormula());
      bench.setRootFormula (root);
      assert (subst.size() == 0);
      writeSimplifiedFormula (bench, subst);
      bench.release();
      assert (mgr.numberOfNodes() == 0);

      /* read in merged formula */
      Parser.init(mgr, new BufferedReader (new FileReader(this.tmpName)));
      bench = Parser.parseSMTBenchmark();
      Parser.reset();
      proc = rt.exec(this.cmdArray);
      if (proc.waitFor() != this.goldenExitCode) {
        bench.release();
        assert (mgr.numberOfNodes() == 0);
        printlnVerbosityMsg ("merging assumptions(s) failed: " + 
                             "exit code is different", 0);
        printlnVerbosityMsg ("see temporary file \"" + this.tmpName + 
                             "\" for merged formula", 0);
        System.exit(0);
      } else  {
        printlnVerbosityMsg ("merging assumptions(s) succeeded", 0);
      }
 
    }

    printlnVerbosityMsg ("starting delta-debugging", 1);

    printlnVerbosityMsg ("\ntrying to find new root in boolean layer", 2);

    /* try to set new root */
    do {
      simplified = false;
      root = bench.getRootFormula();
      assert (root != null);
      assert (root.getType() == SMTTypeManager.booleanType);
      children = root.getChildren();
      if (children != null){
        for (SMTNode child: children){
          if (child.getType() == SMTTypeManager.booleanType){
            if (verbosity >= 3) {
              msg = "trying new root " + child.getKind() + ":";
              System.out.format ("%-" + offset + "s", msg);
            }

            subst.put(root.getId(), child);
            newBench = checkSimplifiedFormula (mgr, bench, subst, queue);
            subst.clear ();

            if (newBench != null) {
              bench.release();
              bench = newBench;
              simplified = true;
              break;
            }
          }
        }
      }
    } while (simplified);

    printlnVerbosityMsg ("\ntrying to simplify sub-terms and sub-formulas", 2);

    /* further simplifications */
    skip = new HashMap<Long,SMTNode>();
    root = bench.getRootFormula ();
    queue.add (root);
    children = root.getChildren();
    if (children != null){
      for (SMTNode child : children)
        queue.add(child);
    }

    cache = new HashSet<SMTNode>();
    while ((cur = queue.poll()) != null){

      if (cur.getRefs() == 0)
        continue;

      if (skip.get(cur.getId()) != null)
        continue;

      cache.clear();

      /* try to set term resp. formula to zero resp. false */
      if ((zero = mgr.zero(cur.getType())) != null){
        assert (zero.getType() == cur.getType());

        if (zero == cur) {
          mgr.release (zero);
          continue;
        }

        if (verbosity >= 3) {
          if (cur.getType() == SMTTypeManager.booleanType)
            msg = "trying to substitute " + cur.getKind() + " by FALSE:";
          else
            msg = "trying to substitute " + cur.getKind() + " by ZERO:";
          System.out.format ("%-" + offset + "s", msg);
        }

        subst.put(cur.getId(), zero);
        newBench = checkSimplifiedFormula (mgr, bench, subst, queue);
        mgr.release (zero);
        subst.clear ();

        if (newBench != null) {
          bench.release();
          bench = newBench;
          continue;
        }

        cache.add (zero);
      }

      /* try to set term resp. formula to one resp. true */
      if ((one = mgr.one(cur.getType())) != null){
        assert (one.getType() == cur.getType());

        if (one == cur) {
          mgr.release (one);
          continue;
        }

        if (verbosity >= 3) {
          if (cur.getType() == SMTTypeManager.booleanType)
            msg = "trying to substitute " + cur.getKind() + " by TRUE:";
          else
            msg = "trying to substitute " + cur.getKind() + " by ONE:";
          System.out.format ("%-" + offset + "s", msg);
        }

        subst.put(cur.getId(), one);
        newBench = checkSimplifiedFormula (mgr, bench, subst, queue);
        mgr.release (one);
        subst.clear ();

        if (newBench != null) {
          bench.release();
          bench = newBench;
          continue;
        }

        cache.add (one);
      }
      
      children = cur.getChildren();
      if (children != null)
        for (SMTNode child : children)
          queue.add(child);

      simplified = false;

      /* We try to skip unary operator chains and STORE chains */
      children = cur.getChildren();
      if (children != null){
        for (SMTNode child : children){
          if (cur.getType() == child.getType()){
            assert (subst.size() == 0);

            if (child.getKind().getArity() != 1 &&
                child.getKind() !=  SMTNodeKind.STORE)
              continue;

            do {
              tmp = child.getChildren()[0];
              if (tmp.getType() == cur.getType())
                child = tmp;
              else
               break;
            } while (child.getKind().getArity() == 1 || 
                     child.getKind() == SMTNodeKind.STORE);

            assert (child.getType() == cur.getType());

            if (!cache.contains(child)) {
              if (verbosity >= 3) {
                msg = "trying to substitute " + cur.getKind() + 
                      " by successor " + child.getKind() + ":";
                System.out.format ("%-" + offset + "s", msg);
              }

              subst.put(cur.getId(), child);
              newBench = checkSimplifiedFormula (mgr, bench, subst, queue);
              subst.clear ();

              if (newBench != null) {
                bench.release();
                bench = newBench;
                simplified = true;
                break;
              }

              cache.add(child);
            }
          }
        }
      }

      if (simplified)
        continue;

      /* try to substitute node by one of its direct children */
      assert (!simplified);
      children = cur.getChildren();
      if (children != null){
        for (SMTNode child : children){
          if (cur.getType() == child.getType()){
            assert (subst.size() == 0);

            if (!cache.contains(child)) {
              if (verbosity >= 3) {
                msg = "trying to substitute " + cur.getKind() + 
                      " by child " + child.getKind() + ":";
                System.out.format ("%-" + offset + "s", msg);
              }

              subst.put(cur.getId(), child);
              newBench = checkSimplifiedFormula (mgr, bench, subst, queue);
              subst.clear ();

              if (newBench != null) {
                bench.release();
                bench = newBench;
                simplified = true;
                break;
              }

              cache.add(child);
            }
          }
        }
      }

      if (simplified)
        continue;

      /* no success, insert into skip table to prevent additional tries */
      skip.put(cur.getId(), cur);
    }
    skip.clear();
    cache.clear();

    /* trying to find new root in the remaining formula
     * we do not search in the boolean layer only, but also
     * inside sub-terms, e.g. ite 
     */

    formulas = bench.getFormulas ();
    assert (formulas[0] == bench.getRootFormula ());
    root = formulas[0];
    if (formulas.length > 1)
      printlnVerbosityMsg ("\ntrying to find new root", 2);
    for (int i = formulas.length - 1; i > 0; i--){
      cur = formulas[i];

      if (verbosity >= 3) {
        msg = "trying new root " + cur.getKind() + ":";
        System.out.format ("%-" + offset + "s", msg);
      }

      subst.put(root.getId(), cur);
      newBench = checkSimplifiedFormula (mgr, bench, subst, queue);
      subst.clear ();

      if (newBench != null) {
        bench.release();
        bench = newBench;
        simplified = true;
        break;
      }

    }

    bench.release();

    if (this.verbosity > 0) {
      printlnVerbosityMsg ("\noriginal formula:   " + origSize + " bytes", 1);
      reducedSize = getFileSize (this.outName);
      printVerbosityMsg ("reduced formula:    ", 1);
      System.out.format ("%" + String.valueOf (origSize).length() + 
                  "d bytes (",  reducedSize);
      System.out.format ("%.2f", ((double) reducedSize / origSize) * 100.0);
      printlnVerbosityMsg ("%)", 1);
    }

    assert (mgr.numberOfNodes() == 0);
  }

  private class Interrupter extends TimerTask {

    private Thread t;

    Interrupter (Thread t) {
      assert (t != null);
      this.t = t;
    }

    public void run() { 
      t.interrupt(); 
    }

  }

  private enum SimplifiedResult {
    YES, NO, TIMEOUT
  }

}
