/*  DeltaSMT: Delta-debugger for Satisfiablity Modulo Theories (SMT) benchmarks.
 *  Copyright (C) 2009  Robert Daniel Brummayer
 *
 *  This file is part of DeltaSMT.
 *
 *  DeltaSMT is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DeltaSMT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.util.*;
import java.math.*;

public class Parser
{

  private static SMTBenchmark g_benchmark;

  /* last recognized token */
  private static Token g_t;

  /* look ahead token */
  private static Token g_la;

  /* shortcut to kind attribute of lookahead token */
  private static TokenKind g_sym;

  private static HashMap<String,SMTNode> g_fletVars;

  private static HashMap<String,SMTNode> g_letVars;

  private static HashMap<String,List<SMTType>> g_sigExtrafuns;

  private static HashMap<String,List<SMTType>> g_sigExtrapreds;

  private static HashMap<String,UType> g_extraSorts;

  private static ArrayList<SMTNode> g_assumptions;

  private static boolean g_parseAssumptions;
  
  private static int g_numAssumptions;

  private static SMTNodeManager g_mgr;


  public static void init (SMTNodeManager mgr, BufferedReader in) 
    throws FileNotFoundException {
    assert (in != null);
    reset ();
    Scanner.init(in);
    g_benchmark = new SMTBenchmark(mgr);
    g_sigExtrafuns = new HashMap<String, List<SMTType>>();
    g_sigExtrapreds = new HashMap<String, List<SMTType>>();
    g_extraSorts = new HashMap<String,UType>();
    g_assumptions = new ArrayList<SMTNode>();
    g_parseAssumptions = false;
    g_numAssumptions = 0;
    g_mgr = mgr;
  }

  public static void reset (){
    Scanner.reset();
    g_benchmark = null;
    g_t = null;
    g_la = null;
    g_sym = null;
    g_fletVars = null;
    g_letVars = null;
    g_sigExtrafuns = null;
    g_sigExtrapreds = null;
    g_extraSorts = null;
    g_assumptions = null;
    g_parseAssumptions = false;
    g_numAssumptions = 0;
    g_mgr = null; 
  }

  /* reads one symbol ahead */
  private static void scan() throws ParserException {
    g_t = g_la;
    try {
      g_la = Scanner.next();
    } catch (ScannerException se){
      throw new ParserException(se.getMessage());
    }
    g_sym = g_la.kind;
  }

  private static void parseError (String message) throws ParserException {
    assert (message != null);
    throw new ParserException (g_la.line + ":" + g_la.col + " " + message);
  }

  /* checks symbol and reads ahead */
  private static void check (TokenKind expected) throws ParserException {
    assert (expected != null);

    if (g_sym != expected && g_sym == TokenKind.EOF){
      throw new ParserException (g_la.line + ":" + g_la.col + 
                                 " unexpected end of file");
    }

    if (g_sym == expected)
      scan();
    else
      parseError ("expected: " + expected.getString());
  }
  
  private static void setLogic (String logicString) throws ParserException {
    SMTLogic logic;

    assert (logicString != null);

    logic = SMTLogic.stringToLogic.get(logicString);
    if (logic == null)
      throw new ParserException ("unsupported logic: " + logicString);
    g_benchmark.setLogic (logic);
  }

  public static SMTBenchmark parseSMTBenchmark () throws ParserException {
    scan();
    check (TokenKind.LPAR);
    check (TokenKind.BENCHMARK);
    check (TokenKind.IDENT);
    g_benchmark.setName(g_t.str);
    while (TokenKind.isBenchAttribute(g_sym)){
      switch (g_sym){
        case A_LOGIC:
          if (g_benchmark.getLogic() != null)
            parseError ("redeclaration of logic");
          scan();
          check (TokenKind.IDENT);
          setLogic (g_t.str);
          break;
        case A_ASSUMPTION:
          if (g_benchmark.getLogic() == null)
            parseError ("logic must be declared before assumption");
          scan();
          g_fletVars = new HashMap<String,SMTNode>();
          g_letVars = new HashMap<String,SMTNode>();
          g_parseAssumptions = true;
          g_assumptions.add (parseFormula());
          g_numAssumptions++;
          g_parseAssumptions = false;
          for (SMTNode n : g_fletVars.values())
            g_mgr.release(n);
          g_fletVars = null;
          for (SMTNode n: g_letVars.values())
            g_mgr.release(n);
          g_letVars = null;
          break;
        case A_FORMULA:
          if (g_benchmark.getRootFormula() != null)
            parseError ("only one formula allowed");
          if (g_benchmark.getLogic() == null)
            parseError ("logic must be declared before formula");
          scan();
          g_fletVars = new HashMap<String,SMTNode>();
          g_letVars = new HashMap<String,SMTNode>();
          g_benchmark.setRootFormula(parseFormula());
          for (SMTNode n : g_fletVars.values())
            g_mgr.release(n);
          g_fletVars = null;
          for (SMTNode n: g_letVars.values())
            g_mgr.release(n);
          g_letVars = null;
          break;
        case A_STATUS:
          if (g_benchmark.getStatus() != null)
            parseError ("redeclaration of status");
          scan ();
          switch (g_sym){
            case SAT:
              g_benchmark.setStatus (SMTBenchmarkStatus.SAT);
              break;
            case UNSAT:
              g_benchmark.setStatus (SMTBenchmarkStatus.UNSAT);
              break;
            case UNKNOWN:
              g_benchmark.setStatus (SMTBenchmarkStatus.UNKNOWN);
              break;
            default:
              parseError ("expected status");
          }
          scan();
          break;
        case A_EXTRASORTS:
          if (g_benchmark.getLogic() == null)
            parseError ("logic must be declared before extrasorts");
          scan();
          parseExtraSorts();
          break;
        case A_EXTRAFUNS:
          if (g_benchmark.getLogic() == null)
            parseError ("logic must be declared before extrafuns");
          scan();
          parseExtrafuns();
          break;
        case A_EXTRAPREDS:
          if (g_benchmark.getLogic() == null)
            parseError ("logic must be declared before extrapreds");
          scan();
          parseExtrapreds();
          break;
        case A_NOTES:
          scan();
          check (TokenKind.STRING);
          break;
        case A_SOURCE: 
        case A_DIFFICULTY: 
        case A_CATEGORY: 
          scan();
          check (TokenKind.USER_VALUE);
          break;
        default:
          throw new ParserException ("Internal error");
      }
    }
    check (TokenKind.RPAR);
    check (TokenKind.EOF);

    g_sigExtrafuns = null;
    g_sigExtrapreds = null;

    g_benchmark.setExtraSorts(g_extraSorts);
    g_extraSorts = null;

    g_benchmark.setAssumptions(g_assumptions);
    g_assumptions = null;
 
    return g_benchmark;
  } 

  private static SMTNode parsePropAtom () throws ParserException {
    SMTNode result;
    List<SMTType> types;
    result = null;
     switch (g_sym){
        case TRUE:
          scan();
          result = SMTNodeManager.TRUE;
          break;
        case FALSE:
          scan();
          result = SMTNodeManager.FALSE;
          break;
        case FVAR:
          if (g_parseAssumptions)
            result = g_fletVars.get("assumption__" + g_numAssumptions + 
                                    "__" + g_la.str);
          else
            result = g_fletVars.get(g_la.str);
          if (result == null)
            parseError ("fvar has not been declared before: $" + g_la.str);
          result.incRefCounter();
          scan();
          break;
        case IDENT:
          types = g_sigExtrapreds.get(g_la.str);
          if (types != null){
            assert (types.size() >= 0);
            if (types.size() > 0)
              parseError ("expected extrapred variable");
            result = g_mgr.extrapred (g_la.str, null);
            scan();
          } else {
            parseError ("extrapred variable has not been declared");
          }
          break;
        default:
          throw new ParserException ("Internal error");
     }
     return result;
  }

  private static SMTNode parseDistinct () throws ParserException,  
    TypeException {
    ArrayList<SMTNode> list;
    SMTNode []nodes;
    SMTNode result;

    assert (g_sym == TokenKind.DISTINCT);

    scan();
    list = new ArrayList<SMTNode>();
    list.add(parseTerm());
    list.add(parseTerm());
    while (g_sym != TokenKind.RPAR)
      list.add(parseTerm());
    nodes = new SMTNode[list.size()];
    list.toArray(nodes);
    result = g_mgr.distinct (nodes);
    for (SMTNode node : list)
      g_mgr.release (node);
    return result;
  }

  private static SMTNode parseFormula () throws ParserException {
    SMTNode result, n0, n1, n2;
    String name;
    List<SMTType> types;
    ArrayList<SMTNode> list;
    SMTNode []nodes;
    Token fun = null;
    result = null;

    if (g_sym == TokenKind.TRUE || g_sym == TokenKind.FALSE || 
        g_sym == TokenKind.FVAR || g_sym == TokenKind.IDENT)
      return parsePropAtom();

    try {
      switch (g_sym){
        case LPAR:
          scan();
          fun = g_la;
          switch (g_sym){
            case NOT:
              scan();
              n0 = parseFormula();
              result = g_mgr.not(n0);
              g_mgr.release (n0);
              break;
            case AND:
              scan();
              list = new ArrayList<SMTNode>();
              list.add(parseFormula());
              list.add(parseFormula());
              while (g_sym != TokenKind.RPAR)
                list.add(parseFormula());
              nodes = new SMTNode[list.size()];
              list.toArray(nodes);
              result = g_mgr.and (nodes);
              for (SMTNode node : list)
                g_mgr.release (node);
              break;
            case OR:
              scan();
              list = new ArrayList<SMTNode>();
              list.add(parseFormula());
              list.add(parseFormula());
              while (g_sym != TokenKind.RPAR)
                list.add(parseFormula());
              nodes = new SMTNode[list.size()];
              list.toArray(nodes);
              result = g_mgr.or (nodes);
              for (SMTNode node : list)
                g_mgr.release (node);
              break;
            case IMPLIES:
              scan();
              n0 = parseFormula();
              n1 = parseFormula();
              result = g_mgr.implies(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case XOR:
              scan();
              n0 = parseFormula();
              n1 = parseFormula();
              result = g_mgr.xor(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case IFF:
              scan();
              n0 = parseFormula();
              n1 = parseFormula();
              result = g_mgr.iff(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case IF_THEN_ELSE:
              scan();
              n0 = parseFormula();
              n1 = parseFormula();
              n2 = parseFormula();
              result = g_mgr.if_then_else(n0, n1, n2);
              g_mgr.release (n0);
              g_mgr.release (n1);
              g_mgr.release (n2);
              break;
            case BVULT:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvult(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVULE:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvule(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVUGT:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvugt(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVUGE:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvuge(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVSLT:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvslt(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVSLE:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvsle(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVSGT:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvsgt(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVSGE:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvsge(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case LT:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.lt(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case LE:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.le(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case GT:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.gt(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case GE:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.ge(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case EQ:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.eq(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case DISTINCT:
              result = parseDistinct();
              break;
            case IDENT:
              name = g_la.str;
              types = g_sigExtrapreds.get (g_la.str);
              list = new ArrayList<SMTNode>();
              assert (types != null);
              assert (types.size() >= 0);
              if (types.size() == 0)
                parseError ("extrapred variable not allowed here");
              scan();
              while (g_sym != TokenKind.RPAR)
                list.add(parseTerm());
              if (list.size() != types.size())
                parseError ("number of arguments does not match");
              for (int i = 0; i < types.size(); i++)
                if (types.get(i) != list.get(i).getType())
                  parseError (name + ": type error: expected " + 
                              types.get(i).toString() + " but was " +
                              list.get(i).getType().toString());
              nodes = new SMTNode[list.size()];
              list.toArray(nodes);
              result = g_mgr.extrapred (name, nodes);
              for (SMTNode node : list)
                g_mgr.release (node);
              break;
            case EXISTS:
            case FORALL:
              parseError("quantifiers unsupported");
            case LET: 
              scan();
              check(TokenKind.LPAR);
              check(TokenKind.VAR);
              name = g_t.str;
              result = parseTerm();
              if (g_letVars.get(name) != null)
                parseError ("redeclaration of var unsupported: " + name); 
              if (g_parseAssumptions)
                g_letVars.put ("assumption__" + g_numAssumptions + "__" + 
                               name, result);
              else
                g_letVars.put (name, result);
              check(TokenKind.RPAR);
              result = parseFormula();
              break;
            case FLET: 
              scan();
              check(TokenKind.LPAR);
              check(TokenKind.FVAR);
              name = g_t.str;
              result = parseFormula();
              if (g_fletVars.get(name) != null)
                parseError ("redeclaration of fvar unsupported: " + name);
              if (g_parseAssumptions)
                g_fletVars.put ("assumption__" + g_numAssumptions + "__" + 
                                name, result);
              else
                g_fletVars.put (name, result);
              check(TokenKind.RPAR);
              result = parseFormula();
              break;
            default:
              throw new ParserException (fun.line + ":" + fun.col + 
                                         " expected boolean connective," +
                                         " predicate, let, flet" +
                                         " or quantifier");
          }
          check (TokenKind.RPAR);
          break;
        default:
          parseError("Expected <an_formula>");
      }
      return result;
    } catch (TypeException te) {
      throw new ParserException(fun.line + ":" + fun.col + " " + fun.str + 
                                ": " + te.getMessage());
    }
  }

  private static SMTNode parseBaseTerm () throws ParserException {
    SMTNode result;
    BigInteger value;
    List<SMTType> types;
    switch (g_la.kind){
      case VAR:
        if (g_parseAssumptions)
          result = g_letVars.get("assumption__" + g_numAssumptions + "__" + 
                                 g_la.str);
        else
          result = g_letVars.get(g_la.str);
        if (result == null)
          parseError ("var has not been declared before: " + g_la.str);
        result.incRefCounter();
        scan();
        break;
      case NUMERAL:
        if (g_la.str.charAt(0) == '-')
          parseError("negative integer numerals unsupported in <base_term>");
        switch (g_benchmark.getLogic()){
          case QF_RDL:
          case QF_LRA:
          case QF_NRA:
          case QF_UFRDL:
          case QF_UFLRA:
          case QF_UFNRA:
            result = g_mgr.intconstasreal (new BigInteger(g_la.str));
            break;
          default:
            result = g_mgr.intconst (new BigInteger(g_la.str));
            break;
        }
        scan();
        break;
      case IDENT:
        types = g_sigExtrafuns.get(g_la.str);
        if (types != null){
          assert (types.size() > 0);
          if (types.size() > 1)
            parseError ("expected extrafun variable or bit-vector constant");
          result = g_mgr.extrafun (types.get(0), g_la.str, null);
          scan();
          break;
        }

        if (!g_la.str.matches ("bv(\\d)+"))
          parseError("expected extrafun variable or bit-vector constant");
        value = new BigInteger(g_la.str.substring(2, g_la.str.length()));
        scan();
        if (g_sym != TokenKind.LBRACK)
          parseError("invalid bit-vector constant");
        scan();
        if (g_sym != TokenKind.NUMERAL)
          parseError("invalid bit-vector constant");
        result = g_mgr.bvconst (Integer.parseInt(g_la.str), value);
        scan();
        if (g_sym != TokenKind.RBRACK)
          parseError("invalid bit-vector constant");
        scan();
        break;
     default:
        throw new ParserException ("Internal error");
    }
    return result;
  }

  private static SMTNode parseTerm () throws ParserException {
    SMTNode result, n0, n1, n2;
    SMTNode []nodes;
    ArrayList<SMTNode> list;
    List<SMTType> types;
    int index1, index2;
    Token fun;
    fun = null;
    result = null;
    String name;

    if (g_sym == TokenKind.VAR || g_sym == TokenKind.NUMERAL ||
        g_sym == TokenKind.IDENT)
      return parseBaseTerm ();

    try {
      switch (g_sym){
        case LPAR:
          scan();
          fun = g_la;
          switch (g_sym){
            case BVNOT:
              scan();
              n0 = parseTerm();
              result = g_mgr.bvnot (n0);
              g_mgr.release (n0);
              break;
            case BVNEG:
              scan();
              n0 = parseTerm();
              result = g_mgr.bvneg (n0);
              g_mgr.release (n0);
              break;
            case TILDE:
              scan();
              n0 = parseTerm();
              result = g_mgr.neg (n0);
              g_mgr.release (n0);
              break;
            case EXTRACT:
              scan();
              check (TokenKind.LBRACK);
              check (TokenKind.NUMERAL);
              index1 = Integer.parseInt (g_t.str);
              check (TokenKind.COLON);
              check (TokenKind.NUMERAL);
              index2 = Integer.parseInt (g_t.str);
              if (index2 > index1)
                throw new TypeException ("upper must be greater than lower");
              check (TokenKind.RBRACK);
              n0 = parseTerm();
              if (!(n0.getType() instanceof BVType))
                throw new TypeException ("expected operand to be of BVType, " +
                                         " but was " + n0.getType().toString());
              if (index1 >= ((BVType)n0.getType()).getWidth())
                throw new TypeException ("upper must be less " +
                                         "than bit-width of operand"); 
              result = g_mgr.extract (n0, index1, index2);
              g_mgr.release (n0);
              break;

            case REPEAT:
              scan();
              check (TokenKind.LBRACK);
              check (TokenKind.NUMERAL);
              index1 = Integer.parseInt (g_t.str);
              if (index1 == 0)
                parseError("repeat[0] unsupported");
              check (TokenKind.RBRACK);
              n0 = parseTerm();
              result = g_mgr.repeat (n0, index1);
              g_mgr.release (n0);
              break;
            case ZERO_EXTEND:
              scan();
              check (TokenKind.LBRACK);
              check (TokenKind.NUMERAL);
              index1 = Integer.parseInt (g_t.str);
              check (TokenKind.RBRACK);
              n0 = parseTerm();
              result = g_mgr.zero_extend (n0, index1);
              g_mgr.release (n0);
              break;
            case SIGN_EXTEND:
              scan();
              check (TokenKind.LBRACK);
              check (TokenKind.NUMERAL);
              index1 = Integer.parseInt (g_t.str);
              check (TokenKind.RBRACK);
              n0 = parseTerm();
              result = g_mgr.sign_extend (n0, index1);
              g_mgr.release (n0);
              break;
            case ROTATE_LEFT:
              scan();
              check (TokenKind.LBRACK);
              check (TokenKind.NUMERAL);
              index1 = Integer.parseInt (g_t.str);
              check (TokenKind.RBRACK);
              n0 = parseTerm();
              if (!(n0.getType() instanceof BVType))
                throw new TypeException ("expected operand to be of BVType, " +
                                         " but was " + n0.getType().toString());
              if (index1 >= ((BVType)n0.getType()).getWidth())
                throw new TypeException ("number of rotates must be less " +
                                         "than bit-width of operand"); 
              result = g_mgr.rotate_left (n0, index1);
              g_mgr.release (n0);
              break;
            case ROTATE_RIGHT:
              scan();
              check (TokenKind.LBRACK);
              check (TokenKind.NUMERAL);
              index1 = Integer.parseInt (g_t.str);
              check (TokenKind.RBRACK);
              n0 = parseTerm();
              if (!(n0.getType() instanceof BVType))
                throw new TypeException ("expected operand to be of BVType, " +
                                         " but was " + n0.getType().toString());
              if (index1 >= ((BVType)n0.getType()).getWidth())
                throw new TypeException ("number of rotates must be less " +
                                         "than bit-width of operand"); 
              result = g_mgr.rotate_right (n0, index1);
              g_mgr.release (n0);
              break;
            case BVAND:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvand (n0,n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVNAND:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvnand (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVOR:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvor (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVNOR:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvnor (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVXOR:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvxor (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVXNOR:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvxnor (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVADD:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvadd (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVMUL:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvmul (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVCOMP:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvcomp (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVSHL:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvshl (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVLSHR:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvlshr (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVASHR:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvashr (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVSUB:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvsub (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVUDIV:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvudiv (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVUREM:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvurem (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVSDIV:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvsdiv (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVSREM:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvsrem (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case BVSMOD:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.bvsmod (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case CONCAT:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.concat (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case PLUS:
              scan();
              list = new ArrayList<SMTNode>();
              list.add(parseTerm());
              while (g_sym != TokenKind.RPAR)
                list.add(parseTerm());
              nodes = new SMTNode[list.size()];
              list.toArray(nodes);
              result = g_mgr.add (nodes);
              for (SMTNode node : list)
                g_mgr.release (node);
              break;
            case MINUS:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.sub(n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case TIMES:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.mul (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case DIV:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.div (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case SELECT:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              result = g_mgr.select (n0, n1);
              g_mgr.release (n0);
              g_mgr.release (n1);
              break;
            case ITE:
              scan();
              n0 = parseFormula();
              n1 = parseTerm();
              n2 = parseTerm();
              result = g_mgr.ite (n0, n1, n2);
              g_mgr.release (n0);
              g_mgr.release (n1);
              g_mgr.release (n2);
              break;
            case STORE:
              scan();
              n0 = parseTerm();
              n1 = parseTerm();
              n2 = parseTerm();
              result = g_mgr.store (n0, n1, n2);
              g_mgr.release (n0);
              g_mgr.release (n1);
              g_mgr.release (n2);
              break;
            case IDENT:
              name = g_la.str;
              types = g_sigExtrafuns.get (g_la.str);
              list = new ArrayList<SMTNode>();
              assert (types != null);
              assert (types.size() > 0);
              if (types.size() == 1)
                parseError ("extrafun variable not allowed here");
              scan();
              while (g_sym != TokenKind.RPAR)
                list.add(parseTerm());
              if (list.size() != types.size() - 1)
                parseError ("number of arguments does not match");
              for (int i = 0; i < types.size() - 1; i++)
                if (types.get(i) != list.get(i).getType())
                  parseError (name + ": type error: expected " + 
                              types.get(i).toString() + " but was " +
                              list.get(i).getType().toString());
              nodes = new SMTNode[list.size()];
              list.toArray(nodes);
              result = g_mgr.extrafun (types.get(types.size() - 1), 
                                       name,nodes);
              for (SMTNode node : list)
                g_mgr.release (node);
              break;
            default:
              parseError ("unsupported function symbol: " + g_la.str);
          }
          check (TokenKind.RPAR);
          break;
        default:
          parseError("Expected <an_term>");
      }
      return result;
    } catch (TypeException te){
      throw new ParserException(fun.line + ":" + fun.col + " " + fun.str + 
                                ": " + te.getMessage());
    }
  }

  private static List<SMTType> parseSorts () throws ParserException {
    List<SMTType> types = new ArrayList<SMTType>();
    int width, indexWidth;
    while (g_sym == TokenKind.IDENT) {
      scan();
      if (g_t.str.equals ("BitVec")) {
        check (TokenKind.LBRACK);
        check (TokenKind.NUMERAL);
        width = Integer.parseInt(g_t.str);
        if (width == 0)
          parseError ("bit-width must be greater than zero");
        assert (width > 0);
        types.add (g_mgr.getTypeManager().BVType(width));
        check (TokenKind.RBRACK);
      } else if (g_t.str.equals ("Array")) {
        if (g_sym == TokenKind.LBRACK) {
          scan();
          check (TokenKind.NUMERAL);
          indexWidth = Integer.parseInt(g_t.str);
          if (indexWidth == 0)
            parseError ("bit-width of array index must be greater than zero");
          assert (indexWidth > 0);
          check (TokenKind.COLON);
          check (TokenKind.NUMERAL);
          width = Integer.parseInt(g_t.str);
          if (width == 0)
            parseError ("bit-width of array element must be greater than zero");
          assert (width > 0);
          types.add(g_mgr.getTypeManager().BVArrayType (indexWidth, width));
          check (TokenKind.RBRACK);
        } else {
          switch (g_benchmark.getLogic()) {
            case QF_A:
            case QF_AX:
              types.add(SMTTypeManager.arrayTypeAX);
              break;
            case QF_AUFLIA:
              types.add(SMTTypeManager.arrayTypeIntInt);
              break;
            default:
              parseError ("Internal error");
              break;
          }
        }
      } else if (g_t.str.equals("Int")) {
        types.add(SMTTypeManager.integerType);
      } else if (g_t.str.equals("Real")) {
        types.add(SMTTypeManager.realType);
      } else if (g_t.str.equals("Index")) {
        types.add(SMTTypeManager.indexType);
      } else if (g_t.str.equals("Element")) {
        types.add(SMTTypeManager.elementType);
      } else if (g_extraSorts.get(g_t.str) != null) {
        types.add(g_extraSorts.get (g_t.str));
      } else {
        parseError ("unsupported type in extrafun declaration");
      }
    }
    return types;
  }

  private static void parseExtrafuns () throws ParserException {
    String name;
    List<SMTType> types;
    check (TokenKind.LPAR);
    while (g_sym == TokenKind.LPAR){
      scan();
      check (TokenKind.IDENT);
      name = g_t.str;
      if (g_sigExtrafuns.get(name) != null)
        parseError ("redeclaration of extrafun: " + name);
      types = parseSorts ();
      assert (types != null);
      if (types.size() == 0)
        parseError ("invalid extrafun declaration");
      g_sigExtrafuns.put (name, types);
      check (TokenKind.RPAR);
    }
    check (TokenKind.RPAR);
  }

  private static void parseExtrapreds () throws ParserException {
    String name;
    List<SMTType> types;
    check (TokenKind.LPAR);
    while (g_sym == TokenKind.LPAR){
      scan();
      check (TokenKind.IDENT);
      name = g_t.str;
      if (g_sigExtrafuns.get(name) != null)
        parseError ("redeclaration of extrafun: " + name);
      types = parseSorts ();
      assert (types != null);
      g_sigExtrapreds.put (name, types);
      check (TokenKind.RPAR);
    }
    check (TokenKind.RPAR);
  }

  private static void parseExtraSorts () throws ParserException {
    String name;
    check(TokenKind.LPAR);
    if (g_sym != TokenKind.IDENT)
      parseError ("expecting extrasort ident");
    while (g_sym == TokenKind.IDENT){
      scan();
      name = g_t.str;
      if (g_extraSorts.get(name) != null)
        parseError ("redeclaration of extrasort: " + name);
      g_extraSorts.put(name, g_mgr.getTypeManager().uType(name));
    }
    check(TokenKind.RPAR);
  }

}
