/*  DeltaSMT: Delta-debugger for Satisfiablity Modulo Theories (SMT) benchmarks.
 *  Copyright (C) 2009  Robert Daniel Brummayer
 *
 *  This file is part of DeltaSMT.
 *
 *  DeltaSMT is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DeltaSMT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.util.*;

public class Scanner
{

  private static final char EOF ='\uffff'; /* used to represent EOF */

  private static final char CR = 13;

  private static final char LF = 10;

  private static Reader g_in;

  private static char g_ch;

  private static int g_line;

  private static int g_col;

  private static HashMap<String, TokenKind> g_keywords;

  private static HashMap<String, TokenKind> g_attribs;

  public static void init(Reader in){
    g_in = in;
    g_line = 1;
    g_col = 0;
    nextChar();
    initHashMap();
  }

  public static void reset() {
    g_in = null;
    g_ch = 0;
    g_line = 0;
    g_col = 0;
    g_keywords = null;
    g_attribs = null;
  }

  private static void initHashMap ()
  {
    g_keywords = new HashMap<String, TokenKind>(58);
    g_keywords.put ("true", TokenKind.TRUE);
    g_keywords.put ("false", TokenKind.FALSE);
    g_keywords.put ("not", TokenKind.NOT);
    g_keywords.put ("and", TokenKind.AND);
    g_keywords.put ("or", TokenKind.OR);
    g_keywords.put ("implies", TokenKind.IMPLIES);
    g_keywords.put ("xor", TokenKind.XOR);
    g_keywords.put ("iff", TokenKind.IFF);
    g_keywords.put ("concat", TokenKind.CONCAT);
    g_keywords.put ("select", TokenKind.SELECT);
    g_keywords.put ("extract", TokenKind.EXTRACT);
    g_keywords.put ("bvadd", TokenKind.BVADD);
    g_keywords.put ("bvnot", TokenKind.BVNOT);
    g_keywords.put ("bvmul", TokenKind.BVMUL);
    g_keywords.put ("bvule", TokenKind.BVULE);
    g_keywords.put ("bvand", TokenKind.BVAND);
    g_keywords.put ("bvlshr", TokenKind.BVLSHR);
    g_keywords.put ("bvslt", TokenKind.BVSLT);
    g_keywords.put ("bvult", TokenKind.BVULT);
    g_keywords.put ("bvneg", TokenKind.BVNEG);
    g_keywords.put ("bvsle", TokenKind.BVSLE);
    g_keywords.put ("bvshl", TokenKind.BVSHL);
    g_keywords.put ("bvsub", TokenKind.BVSUB);
    g_keywords.put ("bvsdiv", TokenKind.BVSDIV);
    g_keywords.put ("bvashr", TokenKind.BVASHR);
    g_keywords.put ("bvor", TokenKind.BVOR);
    g_keywords.put ("bvudiv", TokenKind.BVUDIV);
    g_keywords.put ("bvurem", TokenKind.BVUREM);
    g_keywords.put ("bvnand", TokenKind.BVNAND);
    g_keywords.put ("bvnor", TokenKind.BVNOR);
    g_keywords.put ("bvugt", TokenKind.BVUGT);
    g_keywords.put ("bvuge", TokenKind.BVUGE);
    g_keywords.put ("bvsgt", TokenKind.BVSGT);
    g_keywords.put ("bvsge", TokenKind.BVSGE);
    g_keywords.put ("bvcomp", TokenKind.BVCOMP);
    g_keywords.put ("repeat", TokenKind.REPEAT);
    g_keywords.put ("zero_extend", TokenKind.ZERO_EXTEND);
    g_keywords.put ("sign_extend", TokenKind.SIGN_EXTEND);
    g_keywords.put ("rotate_left", TokenKind.ROTATE_LEFT);
    g_keywords.put ("rotate_right", TokenKind.ROTATE_RIGHT);
    g_keywords.put ("bvxor", TokenKind.BVXOR);
    g_keywords.put ("bvsrem", TokenKind.BVSREM);
    g_keywords.put ("bvsmod", TokenKind.BVSMOD);
    g_keywords.put ("bvxnor", TokenKind.BVXNOR);
    g_keywords.put ("distinct", TokenKind.DISTINCT);
    g_keywords.put ("ite", TokenKind.ITE);
    g_keywords.put ("if_then_else", TokenKind.IF_THEN_ELSE);
    g_keywords.put ("store", TokenKind.STORE);
    g_keywords.put ("exists", TokenKind.EXISTS);
    g_keywords.put ("forall", TokenKind.FORALL);
    g_keywords.put ("let", TokenKind.LET);
    g_keywords.put ("flet", TokenKind.FLET);
    g_keywords.put ("theory", TokenKind.THEORY);
    g_keywords.put ("logic", TokenKind.LOGIC);
    g_keywords.put ("benchmark", TokenKind.BENCHMARK);
    g_keywords.put ("sat", TokenKind.SAT);
    g_keywords.put ("unsat", TokenKind.UNSAT);
    g_keywords.put ("unknown", TokenKind.UNKNOWN);

    g_attribs = new HashMap<String, TokenKind>(19);
    g_attribs.put ("extrafuns", TokenKind.A_EXTRAFUNS);
    g_attribs.put ("assumption", TokenKind.A_ASSUMPTION);
    g_attribs.put ("formula", TokenKind.A_FORMULA);
    g_attribs.put ("status", TokenKind.A_STATUS);
    g_attribs.put ("notes", TokenKind.A_NOTES);
    g_attribs.put ("logic", TokenKind.A_LOGIC);
    g_attribs.put ("sorts", TokenKind.A_SORTS);
    g_attribs.put ("funs", TokenKind.A_FUNS);
    g_attribs.put ("preds", TokenKind.A_PREDS);
    g_attribs.put ("definition", TokenKind.A_DEFINITION);
    g_attribs.put ("axioms", TokenKind.A_AXIOMS);
    g_attribs.put ("theory", TokenKind.A_THEORY);
    g_attribs.put ("language", TokenKind.A_LANGUAGE);
    g_attribs.put ("extensions", TokenKind.A_EXTENSIONS);
    g_attribs.put ("extrasorts", TokenKind.A_EXTRASORTS);
    g_attribs.put ("extrapreds", TokenKind.A_EXTRAPREDS);
    g_attribs.put ("source", TokenKind.A_SOURCE);
    g_attribs.put ("category", TokenKind.A_CATEGORY);
    g_attribs.put ("difficulty", TokenKind.A_DIFFICULTY);

  }

  private static void nextChar (){
    int res;
    try {
      res = g_in.read();
      if (res < 0) {
        g_ch = EOF;
      } else {
        g_col++;
        g_ch = (char) res;
      }
      switch (g_ch){
        case CR:
          nextChar();
          break;
        case LF:
          g_line++;
          g_col = 0;
          break;
      }
    } catch (Exception e) {
      g_ch = EOF;
    }
  }

  private static void readArithSymb (Token t) throws ScannerException
  {
    String str;
    StringBuilder builder = new StringBuilder();
    do {
      builder.append(g_ch);
      nextChar();
    } while (g_ch == '=' || g_ch == '<' || g_ch == '>' || g_ch == '&' ||
             g_ch == '@' || g_ch == '#' || g_ch == '+' || g_ch == '-' ||
             g_ch == '*' || g_ch == '/' || g_ch == '/' || g_ch == '%' ||
             g_ch == '|' || g_ch == '~');
    str = builder.toString();
    t.kind = TokenKind.arithStringToTokenKind.get(str);
    if (t.kind == null)
      throw new ScannerException (t.line + ":" + t.col + 
                                  " unsupported symbol: \"" + str + "\"");
  }

  private static String readIdentifierString ()
  {
    StringBuilder builder = new StringBuilder();

    assert (Character.isLetter(g_ch));

    do {
      builder.append(g_ch);
      nextChar();
    } while (Character.isLetterOrDigit(g_ch) || 
             g_ch == '_' || g_ch == '.' || g_ch == '\'');

    return builder.toString();
  }
  
  private static void readIdentifier (Token t)
  {
    TokenKind kind;
    String str;

    assert (Character.isLetter(g_ch));

    str = readIdentifierString();
    kind = g_keywords.get (str);
    if (kind == null){
        t.kind = TokenKind.IDENT;
        t.str = str;
    } else {
      t.kind = kind;
    }
  }

  private static void readVar (Token t) throws ScannerException
  {
    assert (g_ch == '?');
    nextChar();
    if (!Character.isLetter(g_ch))
      throw new ScannerException (g_line + ":" + g_col + 
                                  " identifier expected");
    t.str = readIdentifierString();
    t.kind = TokenKind.VAR;
  }

  private static void readFVar (Token t) throws ScannerException
  {
    assert (g_ch == '$');

    nextChar();
    if (!Character.isLetter(g_ch))
      throw new ScannerException (g_line + ":" + g_col + 
                                  " identifier expected");
    t.str = readIdentifierString();
    t.kind = TokenKind.FVAR;
  }

  private static void readAttrib (Token t) throws ScannerException
  {
    TokenKind kind;
    String str;

    assert (g_ch == ':');

    nextChar();
    if (!Character.isLetter(g_ch)){
      t.kind = TokenKind.COLON;
    } else {
      str = readIdentifierString();
      /* check for special attribute */
      kind = g_attribs.get (str);
      if (kind == null){
        t.str = str;
        t.kind = TokenKind.ATTRIB;
      } else {
        t.kind = kind;
      }
    }
  }


  private static void readNumeral (Token t)
  {
    StringBuilder builder = new StringBuilder();

    assert (Character.isDigit(g_ch));

    if (g_ch == '0') {
      builder.append(g_ch);
      nextChar();
    } else  {
      do {
        builder.append(g_ch);
        nextChar();
      } while (Character.isDigit(g_ch));
    }
    t.str = builder.toString();
    t.kind = TokenKind.NUMERAL;
  }

  /* we save user value content in t.str */
  private static void readUserValue (Token t) throws ScannerException
  {
    StringBuilder builder = new StringBuilder();

    assert (g_ch == '{');

    do {
      nextChar();

      if (g_ch == '\\')
        {
          builder.append(g_ch);
          nextChar();
          builder.append(g_ch);
          nextChar();
        }

      builder.append(g_ch);

      if (g_ch == '{')
        throw new ScannerException (g_line + ":" + g_col + 
                                    " '{' must be escaped in user value content");
    }
    while (g_ch != '}' && g_ch != EOF);

    assert (builder.length() > 0);
    builder.deleteCharAt (builder.length() - 1);

    if (g_ch == EOF)
        throw new ScannerException (g_line + ":" + g_col + 
                                    " unexpected end of file");

    t.str = builder.toString();
    t.kind = TokenKind.USER_VALUE;
    nextChar();
  }

  /* we save string content in t.str */
  private static void readString (Token t) throws ScannerException
  {
    StringBuilder builder = new StringBuilder();

    assert (g_ch == '\"');

    do {
      nextChar();

      if (g_ch == '\\')
        {
          builder.append(g_ch);
          nextChar();
          builder.append(g_ch);
          nextChar();
        }

      builder.append(g_ch);

    }
    while (g_ch != '\"' && g_ch != EOF);

    assert (builder.length() > 0);
    builder.deleteCharAt (builder.length() - 1);

    if (g_ch == EOF)
        throw new ScannerException (g_line + ":" + g_col + 
                                    " unexpected end of file");

    t.str = builder.toString();
    t.kind = TokenKind.STRING;
    nextChar();
  }

  public static Token next () throws ScannerException {
    while (Character.isWhitespace(g_ch))
      nextChar();

    /* skip comment */
    if (g_ch == ';')
      {
        do {
          nextChar();
        } while (g_ch != LF && g_ch != EOF);
        nextChar();
      }

    Token result = new Token (TokenKind.NONE, g_line, g_col);
    switch (g_ch){
      case '(':
        result.kind = TokenKind.LPAR;
        nextChar();
        break;
      case ')':
        result.kind = TokenKind.RPAR;
        nextChar();
        break;
      case '[':
        result.kind = TokenKind.LBRACK;
        nextChar();
        break;
      case ']':
        result.kind = TokenKind.RBRACK;
        nextChar();
        break;
      case '?':
        readVar(result);
        break;
      case '$':
        readFVar(result);
        break;
      case ':':
        readAttrib(result);
        break;
      case '=':
      case '<':
      case '>':
      case '&':
      case '@':
      case '#':
      case '+':
      case '-':
      case '*':
      case '/':
      case '%':
      case '|':
      case '~':
        readArithSymb(result);
        break;
      case '{':
        readUserValue(result);
        break;
      case '\"':
        readString(result);
        break;
      case EOF:
        result.kind = TokenKind.EOF;
        break;
      default:
        if (Character.isLetter(g_ch)){
          readIdentifier (result);
        } else if (Character.isDigit(g_ch)){
          readNumeral(result);
        } else {
          throw new ScannerException (g_line + ":" + g_col +
                                      " unexpected character: \'" + 
                                      g_ch + "\'");
        }
        break;
    }

    return result;
  }

}
